# Licensed under a 3-clause BSD style license, see LICENSE.
"""
Module for algorithms and methods used to model distributions.

This module contains in particular:

  * Bayesian Blocks binning algorithm.
"""

#-----------------------------------------------------------------------------
# Import statements
#-----------------------------------------------------------------------------
from __future__ import absolute_import

from .bayesian_blocks import bayesian_blocks
